(* source language ---------------------------------------------------------- *)

type var = string
module SRC = struct

  (* Abstract syntax of arithmetic expressions *)
  type exp =
    | Var of var
    | Add of exp * exp
    | Mul of exp * exp
    | Neg of exp
    | Const of int64

  (* Abstract syntax of commands *)
  type cmd =
    | Skip
    | Assn of var * exp
    | Seq  of cmd * cmd
    | IfNZ of exp * cmd * cmd
    | WhileNZ of exp * cmd

  (*
    X2 := X1 + X2;
    IFNZ X2 THEN
      X1 := X1 + 1
    ELSE
      X2 := X1
    X2 := X2 * X1
  *)
  let example_branch : cmd =
    let x1 = "X1" in
    let x2 = "X2" in
    let vx1 = Var x1 in
    let vx2 = Var x2 in
    Seq(Assn(x1, Add(vx1, vx2)),
        Seq(IfNZ(vx2,
                 Assn(x1, Add(vx1, Const 1L)),
                 Assn(x2, vx1)),
          Assn(x2, Mul(vx2, vx1))
        ))


  (* 
     X1 := 6;
     X2 := 1;
     WhileNZ X1 DO
       X2 := X2 * X1;
       X1 := X1 + (-1);
     DONE
  *)
  let factorial : cmd =
    let x = "X1" in
    let ans = "X2" in
    Seq(Assn(x, Const 6L),
        Seq(Assn(ans, Const 1L),
            WhileNZ(Var x,
                    Seq(Assn(ans, Mul(Var ans, Var x)),
                        Assn(x,   Add(Var x, Const (-1L) ))))))
      
  
end


module IR = struct
  
  type uid = string     (* Unique identifiers for temporaries. *)
  type lbl = string
  
  (* "gensym" -- generate a new unique identifier *)
  let mk_uid : unit -> uid =
    let ctr = ref 0 in
    fun () -> let uid = !ctr in
      ctr := !ctr + 1;
      Printf.sprintf "tmp%d" !ctr
  
  
  (* operands *)
  type opn = 
    | Id of uid
    | Const of int64

  (* binary arithmetic operations *)
  type bop =
    | Add
    | Mul

  (* comparison operations *)
  type cmpop =
    | Eq
    | Lt
  
  (* instructions *)
  (* note that there is no nesting of operations! *)
  type insn =
    | Let of uid * bop * opn * opn
    | Load of uid * var
    | Store of var * opn
    | ICmp of uid * cmpop * opn * opn
  
  type terminator =
    | Ret 
    | Br of lbl                (* unconditional branch *)
    | Cbr of opn * lbl * lbl   (* conditional branch *)

  (* Basic blocks *)
  type block = { insns: insn list; terminator: terminator }

  (* Control Flow Graph: a pair of an entry block and a set labeled blocks *)
  type cfg = block * (lbl * block) list
  
  type program = cfg
  
  (* pretty printing *)
  let pp_uid u = u
  let pp_var x = Printf.sprintf "var%s" x
  
  let pp_opn = function 
    | Id u   -> pp_uid u
    | Const c -> (Int64.to_string c) ^ "L"

  let pp_bop = function
    | Add -> "add"
    | Mul -> "mul"

  let pp_cmpop = function
    | Eq -> "eq"
    | lt -> "lt"
  
  let pp_insn = function
    | Let (u, bop, op1, op2) ->
      Printf.sprintf "let %s = %s %s %s"
        (pp_uid u) (pp_bop bop) (pp_opn op1) (pp_opn op2)
    | Load (u, x) ->
      Printf.sprintf "let %s = load %s"
        (pp_uid u) (pp_var x)
    | Store (x, op) ->
      Printf.sprintf "let _ = store %s %s"
        (pp_opn op) (pp_var x)
    | ICmp (u, cmpop, op1, op2) ->
      Printf.sprintf "let %s = icmp %s %s %s"
        (pp_uid u) (pp_cmpop cmpop) (pp_opn op1) (pp_opn op2)

  let pp_terminator = function
    | Ret -> "  ret ()"
    | Br lbl  -> Printf.sprintf "  br %s" lbl
    | Cbr(op, lbl1, lbl2) -> Printf.sprintf "  cbr %s %s %s" (pp_opn op) lbl1 lbl2

  let pp_block {insns; terminator} =
    (String.concat " in\n" (List.map pp_insn insns)) ^
    (pp_terminator terminator)

  let pp_cfg (entry_block, blocks) =
    (Printf.sprintf "let rec entry () =\n%s" (pp_block entry_block)) ^
    (String.concat "\n\n"
       (List.map (fun (lbl,block) -> Printf.sprintf "and %s () =\n%s" lbl (pp_block block)) blocks))
  
  let pp_program cfg =
    Printf.sprintf "let program ()=\n%s\nin entry ()" (pp_cfg cfg)

  module MLMeaning = struct
    let add = Int64.add
    let mul = Int64.mul
    let load (x : int64 ref) = (!x)
    let store o (x : int64 ref) = x := o
    let icmp cmpop x y = cmpop x y 

    let eq (x : int64) (y : int64) = x = y
    let lt x y = x < y

    let ret x = x

    let cbr cnd lbl1 lbl2 =
      if cnd then lbl1 () else lbl2 ()
    let br lbl = lbl ()
  end
  
end

module Compile = struct
  open SRC
  
  let rec compile_exp (e:exp) : (IR.insn list) * IR.opn =
    let compile_bop bop e1 e2 = 
        let ins1, ret1 = compile_exp e1 in
        let ins2, ret2 = compile_exp e2 in
        let ret = IR.mk_uid () in
        ins1 @ ins2 @ IR.[Let (ret, bop, ret1, ret2)], IR.Id ret
    in
    begin match e with
      | Var x ->
        let ret = IR.mk_uid () in
        IR.[Load(ret, x)], IR.Id ret
      | Const c -> [], IR.Const c
      | Add(e1, e2) -> compile_bop IR.Add e1 e2
      | Mul(e1, e2) -> compile_bop IR.Mul e1 e2
      | Neg(e1) -> compile_bop IR.Mul e1 (Const(-1L))
    end

  let rec compile_cmd (c:cmd)  =
    failwith "harder than before!"
    
  let compile (c:cmd) : IR.program =
    failwith "todo"
    

end


